@@ -15,8 +15,9 @@ See L<Net::WebSocket::Endpoint::Server>.
 use strict;
 use warnings;
 
-use Call::Context ();
+use Module::Runtime ();
 
+use Net::WebSocket::Defragmenter ();
 use Net::WebSocket::Frame::close ();
 use Net::WebSocket::Frame::ping ();
 use Net::WebSocket::Frame::pong ();
@@ -37,15 +38,41 @@ sub new {
         die "“out” ($opts{'out'}) needs a write() method!";
     }
 
-    my $self = {
-        _fragments => [],
+    my $self;
 
+    my $defragger = Net::WebSocket::Defragmenter->new(
+        parser => $opts{'parser'},
+
+        on_data_frame => $opts{'on_data_frame'},
+
+        on_control_frame => sub {
+            $self->_handle_control_frame(shift);
+        },
+
+        on_protocol_error => sub {
+            ( undef, my $msg ) = @_;
+
+            #For now … there may be some multiplexing extension
+            #that allows some other behavior down the line,
+            #but let’s enforce standard protocol for now.
+            my $err_frame = Net::WebSocket::Frame::close->new(
+                code => 'PROTOCOL_ERROR',
+                reason => $msg,
+                $self->FRAME_MASK_ARGS(),
+            );
+
+            $self->_write_frame($err_frame);
+        },
+    );
+
+    $self = {
         _max_pings => $class->DEFAULT_MAX_PINGS(),
 
         _ping_store => Net::WebSocket::PingStore->new(),
 
+        _defragger => $defragger,
+
         (map { defined($opts{$_}) ? ( "_$_" => $opts{$_} ) : () } qw(
-            parser
             max_pings
 
             on_data_frame
@@ -62,41 +89,23 @@ sub get_next_message {
 
     $self->_verify_not_closed();
 
-    if ( my $frame = $self->{'_parser'}->get_next_frame() ) {
-        if ($frame->is_control_frame()) {
-            $self->_handle_control_frame($frame);
-        }
-        else {
-            if ($self->{'_on_data_frame'}) {
-                $self->{'_on_data_frame'}->($frame);
-            }
+    return $self->{'_defragger'}->get_next_message();
+}
 
-            #Failure cases:
-            #   - continuation without prior fragment
-            #   - non-continuation within fragment
+sub create_message {
+    my ($self, $frame_type, $payload) = @_;
 
-            if ( $frame->get_type() eq 'continuation' ) {
-                if ( !@{ $self->{'_fragments'} } ) {
-                    $self->_got_continuation_during_non_fragment($frame);
-                }
-            }
-            elsif ( @{ $self->{'_fragments'} } ) {
-                $self->_got_non_continuation_during_fragment($frame);
-            }
+    require Net::WebSocket::FrameTypeName;
 
-            if ($frame->get_fin()) {
-                return Net::WebSocket::Message::create_from_frames(
-                    splice( @{ $self->{'_fragments'} } ),
-                    $frame,
-                );
-            }
-            else {
-                push @{ $self->{'_fragments'} }, $frame;
-            }
-        }
-    }
+    my $frame_class = Net::WebSocket::FrameTypeName::get_module($frame_type);
+    Module::Runtime::require_module($frame_class) if !$frame_class->can('new');
 
-    return undef;
+    return Net::WebSocket::Message->new(
+        $frame_class->new(
+            payload => $payload,
+            $self->FRAME_MASK_ARGS(),
+        ),
+    );
 }
 
 sub check_heartbeat {
@@ -105,20 +114,16 @@ sub check_heartbeat {
     my $ping_counter = $self->{'_ping_store'}->get_count();
 
     if ($ping_counter == $self->{'_max_pings'}) {
-        my $close = Net::WebSocket::Frame::close->new(
-            $self->FRAME_MASK_ARGS(),
+        $self->close(
             code => 'POLICY_VIOLATION',
+            reason => "Unanswered ping(s): $ping_counter",
         );
-
-        $self->_write_frame($close);
-
-        $self->{'_closed'} = 1;
     }
 
     my $ping_message = $self->{'_ping_store'}->add();
 
     my $ping = Net::WebSocket::Frame::ping->new(
-        payload_sr => \$ping_message,
+        payload => $ping_message,
         $self->FRAME_MASK_ARGS(),
     );
 
@@ -127,7 +132,7 @@ sub check_heartbeat {
     return;
 }
 
-sub shutdown {
+sub close {
     my ($self, %opts) = @_;
 
     my $close = Net::WebSocket::Frame::close->new(
@@ -136,16 +141,50 @@ sub shutdown {
         reason => $opts{'reason'},
     );
 
-    $self->_write_frame($close);
+    return $self->_close_with_frame($close);
+}
 
-    $self->{'_closed'} = 1;
+sub _close_with_frame {
+    my ($self, $close_frame) = @_;
+
+    $self->_write_frame($close_frame);
+
+    $self->{'_sent_close_frame'} = $close_frame;
+
+    return $self;
+}
+
+*shutdown = *close;
+
+sub is_closed {
+    my ($self) = @_;
+    return $self->{'_sent_close_frame'} ? 1 : 0;
+}
+
+sub received_close_frame {
+    my ($self) = @_;
+    return $self->{'_received_close_frame'};
+}
+
+sub sent_close_frame {
+    my ($self) = @_;
+    return $self->{'_sent_close_frame'};
+}
+
+sub die_on_close {
+    my ($self) = @_;
+
+    $self->{'_no_die_on_close'} = 0;
 
     return;
 }
 
-sub is_closed {
+sub do_not_die_on_close {
     my ($self) = @_;
-    return $self->{'_closed'} ? 1 : 0;
+
+    $self->{'_no_die_on_close'} = 1;
+
+    return;
 }
 
 #----------------------------------------------------------------------
@@ -155,7 +194,7 @@ sub on_ping {
 
     $self->_write_frame(
         Net::WebSocket::Frame::pong->new(
-            payload_sr => \$frame->get_payload(),
+            payload => $frame->get_payload(),
             $self->FRAME_MASK_ARGS(),
         ),
     );
@@ -173,48 +212,10 @@ sub on_pong {
 
 #----------------------------------------------------------------------
 
-sub _got_continuation_during_non_fragment {
-    my ($self, $frame) = @_;
-
-    my $msg = sprintf('Received continuation outside of fragment!');
-
-    #For now … there may be some multiplexing extension
-    #that allows some other behavior down the line,
-    #but let’s enforce standard protocol for now.
-    my $err_frame = Net::WebSocket::Frame::close->new(
-        code => 'PROTOCOL_ERROR',
-        reason => $msg,
-        $self->FRAME_MASK_ARGS(),
-    );
-
-    $self->_write_frame($err_frame);
-
-    die Net::WebSocket::X->create( 'ReceivedBadControlFrame', $msg );
-}
-
-sub _got_non_continuation_during_fragment {
-    my ($self, $frame) = @_;
-
-    my $msg = sprintf('Received %s; expected continuation!', $frame->get_type());
-
-    #For now … there may be some multiplexing extension
-    #that allows some other behavior down the line,
-    #but let’s enforce standard protocol for now.
-    my $err_frame = Net::WebSocket::Frame::close->new(
-        code => 'PROTOCOL_ERROR',
-        reason => $msg,
-        $self->FRAME_MASK_ARGS(),
-    );
-
-    $self->_write_frame($err_frame);
-
-    die Net::WebSocket::X->create( 'ReceivedBadControlFrame', $msg );
-}
-
 sub _verify_not_closed {
     my ($self) = @_;
 
-    die "Already closed!" if $self->{'_closed'};
+    die Net::WebSocket::X->create('EndpointAlreadyClosed') if $self->{'_closed'};
 
     return;
 }
@@ -222,25 +223,37 @@ sub _verify_not_closed {
 sub _handle_control_frame {
     my ($self, $frame) = @_;
 
-    my ($resp_frame, $error, $ignore_sigpipe);
-
     my $type = $frame->get_type();
 
     if ($type eq 'close') {
-        $self->{'_received_close'} = 1;
-        $self->{'_closed'} = 1;
+        if (!$self->{'_sent_close_frame'}) {
+            my ($code, $reason) = $frame->get_code_and_reason();
 
-        my ($code, $reason) = $frame->get_code_and_reason();
+            if ($code > 4999) {
+                my $reason_chunk = substr( $reason, 0, 50 );
+                $reason = "Invalid close code ($code, reason: $reason_chunk) received.";
+                $code = 'PROTOCOL_ERROR';
+            }
 
-        $resp_frame = Net::WebSocket::Frame::close->new(
-            code => $code,
-            reason => $reason,
-            $self->FRAME_MASK_ARGS(),
-        );
+            $self->_close_with_frame(
+                Net::WebSocket::Frame::close->new(
+                    code => $code,
+                    reason => $reason,
+                    $self->FRAME_MASK_ARGS(),
+                ),
+            );
+        }
 
-        $self->_write_frame( $resp_frame );
+        if ($self->{'_received_close_frame'}) {
+            warn sprintf('Extra close frame received! (%v.02x)', $frame->to_bytes());
+        }
+        else {
+            $self->{'_received_close_frame'} = $frame;
+        }
 
-        die Net::WebSocket::X->create('ReceivedClose', $frame);
+        if (!$self->{'_no_die_on_close'}) {
+            die Net::WebSocket::X->create('ReceivedClose', $frame);
+        }
     }
     elsif ( my $handler_cr = $self->can("on_$type") ) {
         $handler_cr->( $self, $frame );
@@ -257,9 +270,9 @@ sub _handle_control_frame {
 }
 
 sub _write_frame {
-    my ($self, $frame, $todo_cr) = @_;
+    my ($self, $frame) = @_;
 
-    return $self->{'_out'}->write($frame->to_bytes(), $todo_cr);
+    return $self->{'_out'}->write($frame->to_bytes());
 }
 
 1;
